def _get_ranges(idx_series):
    """Generates ranges of contiguous values.

    For example, if input idx_series = [10,11,15,19,20,21,22,23,24,30].

    The output ranges would be:

    [
        (10, 11),
        (15, 15),
        (19, 24),
        (30, 30)
    ]

    Args:
        idx_series (pd.Series): a pandas integer index

    Returns:
        (List(tuple)): a list of tuples, where the tuple is the (starting value, ending value) inclusive
    """
    start_idx = idx_series[0]
    last_idx = idx_series[0]
    ranges = []
    for idx in idx_series[1:]:
        if (idx - last_idx) > 1:
            ranges.append((start_idx, last_idx))
            start_idx = idx

        last_idx = idx

    ranges.append((start_idx, idx_series[-1]))
    return ranges



import pandas as pd


def _bin_numeric_cols_into_categories(schema, data, num_bins):
    """Transforms dataframe columns into numeric categories so that
    dependence can be calculated.

    Args:
        schema (woodwork.TableSchema): Woodwork typing info for the data.
        data (dict[pd.Series]): Dictionary of Pandas series to use for
            calculating dependence.
        num_bins (int): Determines number of bins to use for converting
            numeric features into categorical.

    Returns:
        None
    """
    for col_name in data.keys():
        column = schema.columns[col_name]
        if (column.is_numeric or column.is_datetime) and len(set(data[col_name])) > 2:
            # bin numeric features to make categories when there are enough values
            data[col_name] = pd.qcut(data[col_name], num_bins, duplicates="drop")
        # convert categories to integers
        new_col = data[col_name]
        if str(new_col.dtype) != "category":
            new_col = new_col.astype("category")
        data[col_name] = new_col.cat.codes



from woodwork.utils import get_valid_mi_types


def _get_valid_mi_columns(dataframe, include_index=False, include_time_index=False):
    """Returns a list of columns from the DataFrame with valid
        Logical Types that support mutual information.

    Args:
        dataframe (pd.DataFrame): Data containing Woodwork typing information
            from which to calculate mutual information.
        include_index (bool): If True, the column specified as the index will be
            included as long as its LogicalType is valid for mutual information calculations.
            If False, the index column will not have mutual information calculated for it.
            Defaults to False.
        include_index (bool): If True, the column specified as the time index will be
            included for mutual information calculations.
            If False, the time index column will not have mutual information calculated for it.
            Defaults to False.

    Returns:
        list: A list of column names that have valid Logical Types that support
        mutual information.
    """
    valid_types = tuple(get_valid_mi_types())
    valid_columns = [
        col_name
        for col_name, col in dataframe.ww.columns.items()
        if isinstance(col.logical_type, valid_types)
    ]
    if not include_index and dataframe.ww.index is not None:
        valid_columns.remove(dataframe.ww.index)
    if not include_time_index and dataframe.ww.time_index is not None:
        valid_columns.remove(dataframe.ww.time_index)

    return valid_columns



from woodwork.serializers.parquet_serializer import ParquetSerializer
from woodwork.serializers.csv_serializer import CSVSerializer
from woodwork.serializers.feather_serializer import FeatherSerializer
from woodwork.serializers.arrow_serializer import ArrowSerializer
from woodwork.serializers.orc_serializer import OrcSerializer
from woodwork.serializers.pickle_serializer import PickleSerializer

from woodwork.serializers.utils import get_serializer



import numpy as np

from woodwork.accessor_utils import _is_dask_dataframe, _is_spark_dataframe


def _get_value_counts(dataframe, ascending=False, top_n=10, dropna=False):
    """Returns a list of dictionaries with counts for the most frequent values in each column (only
        for columns with `category` as a standard tag).


    Args:
        dataframe (pd.DataFrame, dd.DataFrame, ps.DataFrame): Data from which to count values.
        ascending (bool): Defines whether each list of values should be sorted most frequent
            to least frequent value (False), or least frequent to most frequent value (True).
            Defaults to False.

        top_n (int): the number of top values to retrieve. Defaults to 10.

        dropna (bool): determines whether to remove NaN values when finding frequency. Defaults
            to False.

    Returns:
        list(dict): a list of dictionaries for each categorical column with keys `count`
        and `value`.
    """
    val_counts = {}
    valid_cols = [
        col for col, column in dataframe.ww.columns.items() if column.is_categorical
    ]
    data = dataframe[valid_cols]
    is_ks = False
    if _is_dask_dataframe(data):
        data = data.compute()
    if _is_spark_dataframe(data):
        data = data.to_pandas()
        is_ks = True

    for col in valid_cols:
        if dropna and is_ks:
            # Spark categorical columns will have missing values replaced with the string 'None'
            # Replace them with np.nan so dropna work
            datacol = data[col].replace(to_replace="None", value=np.nan)
        else:
            datacol = data[col]
        frequencies = datacol.value_counts(ascending=ascending, dropna=dropna)
        df = frequencies[:top_n].reset_index()
        df.columns = ["value", "count"]
        values = list(df.to_dict(orient="index").values())
        val_counts[col] = values
    return val_counts



import sys
from typing import Any, Callable, Iterable, Union

import numpy as np
import pandas as pd
from importlib_resources import files
from pandas.api import types as pdtypes

import woodwork as ww
from woodwork.config import config
from woodwork.type_sys.utils import _is_categorical_series, col_is_datetime

MAX_INT = sys.maxsize
MIN_INT = -sys.maxsize - 1
Tokens = Iterable[str]

COMMON_WORDS_SET = set(
    word.strip().lower()
    for word in files("woodwork.data").joinpath("1-1000.txt").read_text().split("\n")
    if len(word) > 0
)

NL_delimiters = r"[- \[\].,!\?;\n]"


def categorical_func(series):
    if pdtypes.is_categorical_dtype(series.dtype):
        return True

    if pdtypes.is_string_dtype(series.dtype) and not col_is_datetime(series):
        categorical_threshold = ww.config.get_option("categorical_threshold")

        return _is_categorical_series(series, categorical_threshold)

    if pdtypes.is_float_dtype(series.dtype) or pdtypes.is_integer_dtype(series.dtype):
        numeric_categorical_threshold = ww.config.get_option(
            "numeric_categorical_threshold",
        )
        if numeric_categorical_threshold is not None:
            return _is_categorical_series(series, numeric_categorical_threshold)
        else:
            return False

    return False


def integer_func(series):
    if integer_nullable_func(series) and not series.isnull().any():
        if pdtypes.is_object_dtype(series.dtype):
            return True
        return all(series.mod(1).eq(0))
    return False


def integer_nullable_func(series):
    if pdtypes.is_integer_dtype(series.dtype):
        threshold = ww.config.get_option("numeric_categorical_threshold")
        if threshold is not None:
            return not _is_categorical_series(series, threshold)
        else:
            return True
    elif pdtypes.is_float_dtype(series.dtype):

        def _is_valid_int(value):
            return value >= MIN_INT and value <= MAX_INT and value.is_integer()

        if not series.isnull().any():
            return False
        series_no_null = series.dropna()
        return all([_is_valid_int(v) for v in series_no_null])
    elif pdtypes.is_object_dtype(series.dtype):
        series_no_null = series.dropna()
        try:
            return series_no_null.map(
                lambda x: (isinstance(x, str) and isinstance(int(x), int)),
            ).all()
        except ValueError:
            return False

    return False


def double_func(series):
    if pdtypes.is_float_dtype(series.dtype):
        threshold = ww.config.get_option("numeric_categorical_threshold")
        if threshold is not None:
            return not _is_categorical_series(series, threshold)
        else:
            return True
    elif pdtypes.is_object_dtype(series.dtype):
        series_no_null = series.dropna()
        try:
            # If str and casting to float works, make sure that it isn't just an integer
            return series_no_null.map(
                lambda x: isinstance(x, str) and not float(x).is_integer(),
            ).any()
        except ValueError:
            return False

    return False


def boolean_func(series):
    if boolean_nullable_func(series) and not series.isnull().any():
        return True
    return False


def boolean_nullable_func(series):
    if pdtypes.is_bool_dtype(series.dtype) and not pdtypes.is_categorical_dtype(
        series.dtype,
    ):
        return True
    elif pdtypes.is_object_dtype(series.dtype):
        series_no_null = series.dropna()
        try:
            series_no_null_unq = set(series_no_null)
            if series_no_null_unq in [
                {False, True},
                {True},
                {False},
            ]:
                return True
            series_lower = set(str(s).lower() for s in set(series_no_null))
            if series_lower in [
                set(boolean_list)
                for boolean_list in config.get_option("boolean_inference_strings")
            ]:
                return True
        except (
            TypeError
        ):  # Necessary to check for non-hashable values because of object dtype consideration
            return False
    elif pdtypes.is_integer_dtype(series.dtype) and len(
        config.get_option("boolean_inference_ints"),
    ):
        series_unique = set(series)
        if series_unique == set(config.get_option("boolean_inference_ints")):
            return True
    return False


def datetime_func(series):
    if col_is_datetime(series):
        return True
    return False


def timedelta_func(series):
    if pdtypes.is_timedelta64_dtype(series.dtype):
        return True
    return False


def num_common_words(wordlist: Union[Tokens, Any]) -> float:
    if not isinstance(wordlist, Iterable):
        return np.nan
    num_common_words = 0
    for x in wordlist:
        if x.lower() in COMMON_WORDS_SET:
            num_common_words += 1
    return num_common_words


def natural_language_func(series):
    tokens = series.astype("string").str.split(NL_delimiters)
    mean_num_common_words = np.nanmean(tokens.map(num_common_words))

    return (
        mean_num_common_words > 1.14
    )  # determined through https://github.com/alteryx/nl_inference


class InferWithRegex:
    def __init__(self, get_regex: Callable[[], str]):
        self.get_regex = get_regex

    def __call__(self, series: pd.Series) -> bool:
        series = series.dropna()
        regex = self.get_regex()

        # Includes a check for object dtypes
        if not (
            pdtypes.is_categorical_dtype(series.dtype)
            or pdtypes.is_string_dtype(series.dtype)
        ):
            return False

        try:
            series_match_method = series.str.match
        except (AttributeError, TypeError):
            # This can happen either when the inferred dtype for a series is not
            # compatible with the pandas string API (AttributeError) *or* when the
            # inferred dtype is not compatible with the string API `match` method
            # (TypeError)
            return False
        matches = series_match_method(pat=regex)

        return matches.sum() == len(matches)


email_address_func = InferWithRegex(
    lambda: ww.config.get_option("email_inference_regex"),
)
phone_number_func = InferWithRegex(
    lambda: ww.config.get_option("phone_inference_regex"),
)
postal_code_func = InferWithRegex(
    lambda: ww.config.get_option("postal_code_inference_regex"),
)
url_func = InferWithRegex(lambda: ww.config.get_option("url_inference_regex"))
ip_address_func = InferWithRegex(
    lambda: (
        "("
        + ww.config.get_option("ipv4_inference_regex")
        + "|"
        + ww.config.get_option("ipv6_inference_regex")
        + ")"
    ),
)



import pandas as pd


def _clean_timeseries(observed_ts: pd.Series) -> pd.Series:
    """Clean and Normalize time_series for subsequent processing. The following is performed on the time_series:

    - index_reset
    - duplicates dropped
    - na values dropped

    Args:
        observed_ts (Series): The time_series to normalize.

    Returns:
        (Series): The normalized time_series
    """

    observed_ts = observed_ts.reset_index(drop=True)
    observed_ts = observed_ts.drop_duplicates()
    observed_ts = observed_ts.dropna()
    return observed_ts



from typing import Callable

import pandas as pd

from woodwork.accessor_utils import _is_dask_series, _is_spark_series
from woodwork.logical_types import (
    URL,
    Address,
    Age,
    AgeFractional,
    AgeNullable,
    Boolean,
    BooleanNullable,
    Categorical,
    CountryCode,
    CurrencyCode,
    Datetime,
    Double,
    EmailAddress,
    Filepath,
    Integer,
    IntegerNullable,
    IPAddress,
    LatLong,
    LogicalType,
    NaturalLanguage,
    Ordinal,
    PersonFullName,
    PhoneNumber,
    PostalCode,
    SubRegionCode,
    Timedelta,
    Unknown,
    _replace_nans,
)
from woodwork.type_sys.inference_functions import (
    boolean_func,
    boolean_nullable_func,
    categorical_func,
    datetime_func,
    double_func,
    email_address_func,
    integer_func,
    integer_nullable_func,
    ip_address_func,
    natural_language_func,
    phone_number_func,
    postal_code_func,
    timedelta_func,
    url_func,
)

DEFAULT_INFERENCE_FUNCTIONS = {
    Address: None,
    Age: None,
    AgeFractional: None,
    AgeNullable: None,
    Boolean: boolean_func,
    BooleanNullable: boolean_nullable_func,
    Categorical: categorical_func,
    CountryCode: None,
    CurrencyCode: None,
    Datetime: datetime_func,
    Double: double_func,
    EmailAddress: email_address_func,
    Filepath: None,
    PersonFullName: None,
    Integer: integer_func,
    IntegerNullable: integer_nullable_func,
    IPAddress: ip_address_func,
    LatLong: None,
    NaturalLanguage: natural_language_func,
    Ordinal: None,
    PhoneNumber: phone_number_func,
    PostalCode: postal_code_func,
    SubRegionCode: None,
    Timedelta: timedelta_func,
    URL: url_func,
    Unknown: None,
}

# (ParentType, ChildType)
DEFAULT_RELATIONSHIPS = [
    (BooleanNullable, Boolean),
    (Categorical, CountryCode),
    (Categorical, CurrencyCode),
    (Categorical, Ordinal),
    (Categorical, PostalCode),
    (Categorical, SubRegionCode),
    (Double, AgeFractional),
    (Integer, Age),
    (IntegerNullable, AgeNullable),
    (IntegerNullable, Integer),
    (Unknown, EmailAddress),
    (Unknown, IPAddress),
    (Unknown, PhoneNumber),
    (Unknown, PostalCode),
    (Unknown, Timedelta),
    (Unknown, URL),
]

DEFAULT_TYPE = Unknown

INFERENCE_SAMPLE_SIZE = 100_000


class TypeSystem(object):
    def __init__(
        self,
        inference_functions=None,
        relationships=None,
        default_type=DEFAULT_TYPE,
    ):
        """Create a new TypeSystem object. LogicalTypes that are present in the keys of
        the inference_functions dictionary will be considered registered LogicalTypes.

        Args:
            inference_functions (dict[LogicalType->func], optional): Dictionary mapping LogicalTypes
                to their corresponding type inference functions. If None, only the default LogicalType
                will be registered without an inference function.
            relationships (list, optional): List of tuples, each with two elements, specifying parent-child
                relationships between logical types. The first element should be the parent LogicalType. The
                second element should be the child LogicalType. If not specified, will default to an empty list
                indicating all types should be considered root types with no children.
            default_type (LogicalType, optional): The default LogicalType to use if no inference matches are
                found. If not specified, will default to the built-in Unknown LogicalType.
        """
        self.default_type = default_type
        if inference_functions:
            self.inference_functions = inference_functions.copy()
            if self.default_type not in self.inference_functions:
                self.inference_functions[self.default_type] = None
        else:
            self.inference_functions = {self.default_type: None}

        if relationships:
            self.relationships = relationships.copy()
        else:
            self.relationships = []

        # Store initial values for resetting
        self._default_inference_functions = self.inference_functions.copy()
        self._default_relationships = self.relationships.copy()
        self._default_type = self.default_type

    def add_type(
        self,
        logical_type: LogicalType,
        inference_function: Callable = None,
        parent: LogicalType = None,
        treatment: str = None,
    ):
        """Add a new LogicalType to the TypeSystem, optionally specifying the corresponding inference function and a
        parent type.

        Args:
            logical_type (LogicalType): The new LogicalType to add.
            inference_function (func, optional): The inference function to use for inferring the given LogicalType.
                Defaults to None. If not specified, this LogicalType will never be inferred.
            parent (LogicalType, optional): The parent LogicalType, if applicable. Defaults to None. If not specified,
                this type will be considered a root type with no parent.
            treatment (string, optional): If the specified LogicalType already exists, determine the logic that should
                be applied. Options are "replace", "ignore", or None (default). "replace" will unregister the present
                LogicalType and replace it with the one that was passed. "ignore" will not register the passed
                LogicalType if it already exists. The default value of None will raise an error.
        """
        if isinstance(parent, str):
            parent = self.str_to_logical_type(parent)
        if not (
            isinstance(treatment, type(None))
            or (
                isinstance(treatment, str)
                and treatment.lower() in ["replace", "ignore"]
            )
        ):
            raise ValueError(
                "The parameter treatment can only take on the values 'replace', 'ignore', and None.",
            )
        self._validate_type_input(
            logical_type=logical_type,
            inference_function=inference_function,
            parent=parent,
        )

        registered_ltype_names = {ltype.__name__ for ltype in self.registered_types}
        ltype_is_registered = logical_type.__name__ in registered_ltype_names
        if ltype_is_registered and treatment is None:
            raise ValueError(
                f"Logical Type with name {logical_type.__name__} is already present in the Type System. Please rename the LogicalType or remove existing one.",
            )
        elif ltype_is_registered and treatment.lower() == "replace":
            self.remove_type(logical_type=logical_type)
        if not ltype_is_registered or treatment.lower() == "replace":
            self.update_inference_function(logical_type, inference_function)
            if parent:
                self.update_relationship(logical_type, parent)

    def remove_type(self, logical_type: LogicalType, treatment: str = None):
        """Remove a logical type from the TypeSystem. Any children of the remove type will have their parent
        set to the parent of the removed type.

        Args:
            logical_type (LogicalType): The LogicalType to remove.
            treatment (str, optional): If the specified logical type doesn't exist, determine the logic that should
                be applied. Options are "ignore" and None (default). "ignore" will not raise an error if the passed
                LogicalType doesn't exists. The default value of None will raise an error.
        """
        if isinstance(logical_type, str):
            logical_type = self.str_to_logical_type(logical_type)
        if not (
            isinstance(treatment, type(None))
            or (isinstance(treatment, str) and treatment.lower() in ["ignore"])
        ):
            raise ValueError(
                "The parameter treatment can only take on the values 'ignore', and None.",
            )
        self._validate_type_input(logical_type=logical_type)
        # Remove the inference function
        if logical_type == self.default_type:
            raise ValueError("Default LogicalType cannot be removed")

        registered_ltype_names = {ltype.__name__ for ltype in self.registered_types}
        ltype_is_registered = logical_type.__name__ in registered_ltype_names
        if ltype_is_registered:
            self.inference_functions.pop(logical_type)

            # If the removed type had children we need to update them
            children = self._get_children(logical_type)
            if children:
                parent = self._get_parent(logical_type)
                for child in children:
                    self.update_relationship(child, parent)

            # Rebuild the relationships list to remove any reference to the removed type
            self.relationships = [
                rel for rel in self.relationships if logical_type not in rel
            ]
        else:
            if treatment is None or treatment.lower() != "ignore":
                raise ValueError(
                    f"Logical Type with name {logical_type.__name__} is not present in the Type System.",
                )

    def update_inference_function(self, logical_type, inference_function):
        """Update the inference function for the specified LogicalType.

        Args:
            logical_type (LogicalType): The LogicalType for which to update the inference function.
            inference_function (func): The new inference function to use. Can be set to None to skip
                type inference for the specified LogicalType.
        """
        if isinstance(logical_type, str):
            logical_type = self.str_to_logical_type(logical_type)
        self._validate_type_input(
            logical_type=logical_type,
            inference_function=inference_function,
        )
        self.inference_functions[logical_type] = inference_function

    def update_relationship(self, logical_type, parent):
        """Add or update a relationship. If the specified LogicalType exists in the relationship graph,
        its parent will be updated. If the specified LogicalType does not exist in relationships, the
        relationship will be added.

        Args:
            logical_type (LogicalType): The LogicalType for which to update the parent value.
            parent (LogicalType): The new parent to set for the specified LogicalType.
        """
        if isinstance(logical_type, str):
            logical_type = self.str_to_logical_type(logical_type)
        if isinstance(parent, str):
            parent = self.str_to_logical_type(parent)
        self._validate_type_input(logical_type=logical_type, parent=parent)
        # If the logical_type already has a parent, remove that from the list
        self.relationships = [
            rel for rel in self.relationships if rel[1] != logical_type
        ]
        # Add the new/updated relationship
        self.relationships.append((parent, logical_type))

    def reset_defaults(self):
        """Reset type system to the default settings that were specified at initialization.

        Args:
            None
        """
        self.inference_functions = self._default_inference_functions.copy()
        self.relationships = self._default_relationships.copy()
        self.default_type = self._default_type

    @property
    def registered_types(self):
        """Returns a list of all registered types"""
        return list(self.inference_functions.keys())

    @property
    def root_types(self):
        """Returns a list of all registered types that do not have a parent type"""
        return [
            ltype for ltype in self.registered_types if self._get_parent(ltype) is None
        ]

    def _get_children(self, logical_type):
        """List of all the child types for the given logical type"""
        return [child for parent, child in self.relationships if parent == logical_type]

    def _get_parent(self, logical_type):
        """Get the parent type for the given logical type"""
        for parent, child in self.relationships:
            if child == logical_type:
                return parent
        return None

    def _get_depth(self, logical_type):
        """Get the depth of a type in the relationship graph"""
        depth = 0
        parent = self._get_parent(logical_type)
        while parent:
            depth = depth + 1
            parent = self._get_parent(parent)
        return depth

    def _validate_type_input(
        self,
        logical_type=None,
        inference_function=None,
        parent=None,
    ):
        if logical_type and logical_type not in LogicalType.__subclasses__():
            raise TypeError("logical_type must be a valid LogicalType")

        if inference_function and not callable(inference_function):
            raise TypeError("inference_function must be a function")

        if parent and parent not in self.registered_types:
            raise ValueError("parent must be a valid LogicalType")

    def infer_logical_type(self, series):
        """Infer the logical type for the given series

        Args:
            series (pandas.Series): The series for which to infer the LogicalType.
        """

        def get_random_sample(series_, **kwargs):
            if len(series_) > INFERENCE_SAMPLE_SIZE:
                sampled_series = series_.sample(**kwargs)
                return sampled_series
            else:
                return series_

        kw_args_sampling = {
            "replace": False,
            "random_state": 42,
        }
        if isinstance(series, pd.Series):
            # Special case for series with no valid values
            if series.count() == 0:
                return Unknown()
            kw_args_sampling["n"] = INFERENCE_SAMPLE_SIZE
            series = get_random_sample(series, **kw_args_sampling)
        else:
            # Dask and Spark don't accept the n argument

            # prevent division by zero error
            series_len = len(series)
            if not series_len:
                return Unknown()
            kw_args_sampling["frac"] = INFERENCE_SAMPLE_SIZE / series_len
            if _is_dask_series(series):
                series = get_random_sample(
                    series.head(series_len, npartitions=-1), **kw_args_sampling
                )
            elif _is_spark_series(series):
                series = get_random_sample(series, **kw_args_sampling)
                series = series.to_pandas()
            else:
                raise ValueError(
                    f"Unsupported series type `{type(series)}`",
                )  # pragma: no cover

            # For dask or spark collections, unknown type special case comes
            # *after* head calls to avoid evaluating a potentially large
            # dataset
            if series.count() == 0:
                return Unknown()

        def get_inference_matches(types_to_check, series, type_matches=[]):
            # Since NaturalLanguage isn't inferred by default, make sure to check
            # any children of NaturalLanguage, otherwise they never get evaluated
            check_next = []
            for logical_type in types_to_check:
                inference_func = self.inference_functions.get(logical_type)
                if inference_func and inference_func(series):
                    type_matches.append(logical_type)
                    check_next.extend(self._get_children(logical_type))
                elif not inference_func:
                    check_next.extend(self._get_children(logical_type))
            if len(check_next) > 0:
                get_inference_matches(check_next, series, type_matches)
            return type_matches

        # Don't include NaturalLanguage as we only want to check that if
        # no other matches are found
        types_to_check = [
            ltype for ltype in self.root_types if ltype != NaturalLanguage
        ]
        series_nan_cast = _replace_nans(series)  # Will change dtype
        if series_nan_cast.count() == 0:
            return Unknown()

        type_matches = get_inference_matches(types_to_check, series_nan_cast)
        if len(type_matches) == 0:
            # Check if this is NaturalLanguage, otherwise set
            # type to default type (Unknown). Assume that a column
            # can only be natural language if it is not already a
            # match for another type. Also improves performance by
            # limiting the times the natural language inference function
            # is called.
            if self.inference_functions.get(
                NaturalLanguage,
            ) and self.inference_functions[NaturalLanguage](series):
                logical_type = NaturalLanguage
            else:
                logical_type = self.default_type
        elif len(type_matches) == 1:
            # If we match only one type, return it
            logical_type = type_matches[0]
        else:
            # If multiple matches, get the most specific one. If multiple
            # matches have the same level of specificity, the first
            # match found at that level will be returned
            if (
                Categorical in type_matches or Double in type_matches
            ) and IntegerNullable in type_matches:
                best_match = IntegerNullable
            elif Categorical in type_matches and Double in type_matches:
                best_match = Double
            else:
                best_match = type_matches[0]
            best_depth = self._get_depth(best_match)
            for logical_type in type_matches[1:]:
                ltype_depth = self._get_depth(logical_type)
                if ltype_depth > best_depth:
                    best_match = logical_type
                    best_depth = ltype_depth
            logical_type = best_match

        return logical_type()

    def _get_logical_types(self):
        """Returns a dictionary of logical type name strings and logical type classes"""
        # Get snake case strings
        logical_types = {
            logical_type.type_string: logical_type
            for logical_type in self.registered_types
        }
        # Add class name strings
        class_name_dict = {
            logical_type.__name__: logical_type
            for logical_type in self.registered_types
        }
        logical_types.update(class_name_dict)

        return logical_types

    def str_to_logical_type(self, logical_str, params=None, raise_error=True):
        """Helper function for converting a string value to the corresponding logical type object.
        If a dictionary of params for the logical type is provided, apply them."""
        logical_str_lower = logical_str.lower()
        logical_types_dict = {
            ltype_name.lower(): ltype
            for ltype_name, ltype in self._get_logical_types().items()
        }

        if logical_str_lower in logical_types_dict:
            ltype = logical_types_dict[logical_str_lower]
            if params:
                return ltype(**params)
            else:
                return ltype
        elif raise_error:
            raise ValueError("String %s is not a valid logical type" % logical_str)


type_system = TypeSystem(
    inference_functions=DEFAULT_INFERENCE_FUNCTIONS,
    relationships=DEFAULT_RELATIONSHIPS,
    default_type=DEFAULT_TYPE,
)



import numpy as np
import pandas as pd
import pytest

from woodwork.accessor_utils import (
    _is_dask_dataframe,
    _is_dask_series,
    _is_dataframe,
    _is_series,
    _is_spark_dataframe,
    _is_spark_series,
    get_invalid_schema_message,
    init_series,
    is_schema_valid,
)
from woodwork.exceptions import TypeConversionError
from woodwork.logical_types import Categorical, Datetime, LatLong, NaturalLanguage


def test_init_series_valid_conversion_specified_ltype(sample_series):
    if _is_spark_series(sample_series):
        sample_series = sample_series.astype("str")
    else:
        sample_series = sample_series.astype("object")

    series = init_series(sample_series, logical_type="categorical")
    assert series is not sample_series
    correct_dtype = Categorical._get_valid_dtype(type(sample_series))
    assert series.dtype == correct_dtype
    assert isinstance(series.ww.logical_type, Categorical)
    assert series.ww.semantic_tags == {"category"}

    series = init_series(sample_series, logical_type="natural_language")
    assert series is not sample_series
    correct_dtype = NaturalLanguage._get_valid_dtype(type(sample_series))
    assert series.dtype == correct_dtype
    assert isinstance(series.ww.logical_type, NaturalLanguage)
    assert series.ww.semantic_tags == set()


def test_init_series_with_pd_extension_array():
    extension_categories = pd.Categorical([1, 2, 3])
    series = init_series(extension_categories)
    pd_reference_series = init_series(pd.Series([1, 2, 3], dtype="category"))
    assert series.equals(pd_reference_series)
    assert series.ww.logical_type == pd_reference_series.ww.logical_type
    assert series.ww.semantic_tags == pd_reference_series.ww.semantic_tags

    extension_ints = pd.array(np.array([1, 2, 3, 4], dtype="int64"))
    series = init_series(extension_ints)
    pd_reference_series = init_series(pd.Series([1, 2, 3, 4], dtype="Int64"))
    assert series.equals(pd_reference_series)
    assert series.ww.logical_type == pd_reference_series.ww.logical_type
    assert series.ww.semantic_tags == pd_reference_series.ww.semantic_tags


def test_init_series_with_invalid_type(sample_df):
    inputs = [sample_df, 1, "string", None]
    for input_ in inputs:
        error_message = (
            f"Input must be of series type. The current input is of type {type(input_)}"
        )
        with pytest.raises(TypeError, match=error_message):
            init_series(input_)


def test_init_series_with_np_array(sample_series_pandas):
    series = init_series(sample_series_pandas.to_numpy())
    series2 = init_series(
        sample_series_pandas,
    )  # Sample series panda contains ['a','b','c','a']
    assert series.equals(series2)
    assert series.ww.logical_type == series2.ww.logical_type
    assert series.ww.semantic_tags == series2.ww.semantic_tags


def test_init_series_with_multidimensional_np_array():
    input_ = np.array([["a", "b"], ["a", "b"]])
    error_message = f"np.ndarray input must be 1 dimensional. Current np.ndarray is {input_.ndim} dimensional"
    with pytest.raises(ValueError, match=error_message):
        init_series(input_)


def test_init_series_valid_conversion_inferred_ltype(sample_series):
    if _is_spark_series(sample_series):
        sample_series = sample_series.astype("str")
    else:
        sample_series = sample_series.astype("object")

    series = init_series(sample_series)
    assert series is not sample_series
    correct_dtype = Categorical._get_valid_dtype(type(sample_series))
    assert series.dtype == correct_dtype
    assert isinstance(series.ww.logical_type, Categorical)
    assert series.ww.semantic_tags == {"category"}


def test_init_series_with_datetime(sample_datetime_series):
    series = init_series(sample_datetime_series, logical_type="datetime")
    assert series.dtype == "datetime64[ns]"
    assert isinstance(series.ww.logical_type, Datetime)


def test_init_series_with_latlong(latlong_df):
    for column in latlong_df:
        series = latlong_df[column]
        series = init_series(series, logical_type="LatLong")
        assert isinstance(series.ww.logical_type, LatLong)


def test_init_series_all_parameters(sample_series):
    if _is_spark_series(sample_series):
        sample_series = sample_series.astype("str")
    else:
        sample_series = sample_series.astype("object")

    metadata = {"meta_key": "meta_value"}
    description = "custom description"
    origin = "base"
    series = init_series(
        sample_series,
        logical_type="categorical",
        semantic_tags=["custom_tag"],
        metadata=metadata,
        description=description,
        origin=origin,
        use_standard_tags=False,
    )
    assert series is not sample_series
    correct_dtype = Categorical._get_valid_dtype(type(sample_series))
    assert series.dtype == correct_dtype
    assert isinstance(series.ww.logical_type, Categorical)
    assert series.ww.semantic_tags == {"custom_tag"}
    assert series.ww.metadata == metadata
    assert series.ww.description == description
    assert series.ww.origin == origin


def test_init_series_error_on_invalid_conversion(sample_series):
    if _is_dask_series(sample_series):
        pytest.xfail(
            "Dask type conversion with astype does not fail until compute is called",
        )
    if _is_spark_series(sample_series):
        pytest.xfail(
            "Spark allows this conversion, filling values it cannot convert with NaN "
            "and converting dtype to float.",
        )

    error_message = (
        "Error converting datatype for sample_series from type category to type Int64. "
        "Please confirm the underlying data is consistent with logical type IntegerNullable."
    )
    with pytest.raises(TypeConversionError, match=error_message):
        init_series(sample_series, logical_type="integer_nullable")


def test_is_series(sample_df):
    assert _is_series(sample_df["id"])
    assert not _is_series(sample_df)


def test_is_dataframe(sample_df):
    assert _is_dataframe(sample_df)
    assert not _is_dataframe(sample_df["id"])


def test_get_invalid_schema_message(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init(
        name="test_schema",
        index="id",
        logical_types={"id": "Double", "full_name": "PersonFullName"},
    )
    schema = schema_df.ww.schema

    assert get_invalid_schema_message(schema_df, schema) is None
    assert (
        get_invalid_schema_message(sample_df, schema)
        == "dtype mismatch for column id between DataFrame dtype, int64, and Double dtype, float64"
    )

    sampled_df = schema_df.sample(frac=0.3)
    assert get_invalid_schema_message(sampled_df, schema) is None

    dropped_df = schema_df.drop("id", axis=1)
    assert (
        get_invalid_schema_message(dropped_df, schema)
        == "The following columns in the typing information were missing from the DataFrame: {'id'}"
    )

    renamed_df = schema_df.rename(columns={"id": "new_col"})
    assert (
        get_invalid_schema_message(renamed_df, schema)
        == "The following columns in the DataFrame were missing from the typing information: {'new_col'}"
    )


def test_get_invalid_schema_message_dtype_mismatch(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init(
        logical_types={"age": "Categorical", "full_name": "PersonFullName"},
    )
    schema = schema_df.ww.schema

    incorrect_int_dtype_df = schema_df.ww.astype({"id": "Int64"})
    incorrect_bool_dtype_df = schema_df.ww.astype({"is_registered": "Int64"})

    assert (
        get_invalid_schema_message(incorrect_int_dtype_df, schema)
        == "dtype mismatch for column id between DataFrame dtype, Int64, and Integer dtype, int64"
    )
    assert (
        get_invalid_schema_message(incorrect_bool_dtype_df, schema)
        == "dtype mismatch for column is_registered between DataFrame dtype, Int64, and BooleanNullable dtype, boolean"
    )

    # Spark backup dtypes make these checks not relevant
    if not _is_spark_dataframe(sample_df):
        incorrect_str_dtype_df = schema_df.ww.astype(
            {"full_name": "object"},
        )  # wont work for spark
        incorrect_categorical_dtype_df = schema_df.ww.astype(
            {"age": "string"},
        )  # wont work for spark
        assert (
            get_invalid_schema_message(incorrect_str_dtype_df, schema)
            == "dtype mismatch for column full_name between DataFrame dtype, object, and PersonFullName dtype, string"
        )
        assert (
            get_invalid_schema_message(incorrect_categorical_dtype_df, schema)
            == "dtype mismatch for column age between DataFrame dtype, string, and Categorical dtype, category"
        )


def test_get_invalid_schema_message_index_checks(sample_df):
    if not isinstance(sample_df, pd.DataFrame):
        pytest.xfail("Index validation not performed for Dask or Spark DataFrames")

    schema_df = sample_df.copy()
    schema_df.ww.init(
        name="test_schema",
        index="id",
        logical_types={"id": "Double", "full_name": "PersonFullName"},
    )
    schema = schema_df.ww.schema

    different_underlying_index_df = schema_df.copy()
    different_underlying_index_df["id"] = pd.Series([9, 8, 7, 6], dtype="float64")
    assert (
        get_invalid_schema_message(different_underlying_index_df, schema)
        == "Index mismatch between DataFrame and typing information"
    )

    not_unique_df = schema_df.replace({3: 1})
    not_unique_df.index = not_unique_df["id"]
    not_unique_df.index.name = None
    assert (
        get_invalid_schema_message(not_unique_df, schema)
        == "Index column is not unique"
    )

    df = pd.DataFrame(
        {
            "id": pd.Series([5, 4, 3, 2], dtype="float64"),
            "col": pd.Series(["b", "b", "b", "d"], dtype="category"),
        },
    )
    df.ww.init(index="id")
    df_schema = df.ww.schema

    nan_df = df.replace({3: None})
    nan_df["id"] = nan_df["id"].astype("float64")
    nan_df = nan_df.set_index("id", drop=False)
    actual = get_invalid_schema_message(nan_df, df_schema)
    assert actual == "Index contains null values"


def test_is_schema_valid_true(sample_df):
    sample_df.ww.init(index="id", logical_types={"phone_number": "Categorical"})
    copy_df = sample_df.copy()

    assert copy_df.ww.schema is None
    assert is_schema_valid(copy_df, sample_df.ww.schema)


def test_is_schema_valid_false(sample_df):
    sample_df.ww.init()
    schema = sample_df.ww.schema

    invalid_dtype_df = sample_df.astype({"age": "float64"})
    assert not is_schema_valid(invalid_dtype_df, schema)

    missing_col_df = sample_df.drop(columns={"is_registered"})
    assert not is_schema_valid(missing_col_df, schema)


def test_is_dask_dataframe(sample_df_dask):
    assert _is_dask_dataframe(sample_df_dask)
    assert not _is_dask_dataframe(pd.DataFrame())


def test_is_dask_series(sample_series_dask):
    assert _is_dask_series(sample_series_dask)
    assert not _is_dask_series(pd.Series())


def test_is_spark_dataframe(sample_df_spark):
    assert _is_spark_dataframe(sample_df_spark)
    assert not _is_dask_dataframe(pd.DataFrame())


def test_is_spark_series(sample_series_spark):
    assert _is_spark_series(sample_series_spark)
    assert not _is_dask_series(pd.Series())



import pytest

from woodwork.config import Config


@pytest.fixture
def defaults():
    return {
        "option1": "value1",
        "option2": "value2",
    }


@pytest.fixture
def config(defaults):
    return Config(defaults)


def test_init_conifg(defaults):
    config = Config(defaults)
    assert config._data == defaults
    assert config._defaults == defaults
    assert config._data == config._defaults
    assert config._data is not config._defaults


def test_get_option(config):
    assert config.get_option("option1") == "value1"
    assert config.get_option("option2") == "value2"


def test_set_option(config):
    config.set_option("option1", "updated_value")
    assert config.get_option("option1") == "updated_value"
    assert config.get_option("option2") == "value2"


def test_reset_option(config):
    config.set_option("option1", "updated_value")
    assert config.get_option("option1") == "updated_value"
    config.reset_option("option1")
    assert config.get_option("option1") == "value1"


def test_with_options(config):
    old_value_1 = config.get_option("option1")
    old_value_2 = config.get_option("option2")

    with_value_1 = "some_other_value_1"
    with_value_2 = "some_other_value_2"

    with config.with_options(option1=with_value_1, option2=with_value_2):
        assert config.get_option("option1") == with_value_1
        assert config.get_option("option2") == with_value_2

    assert config.get_option("option1") == old_value_1
    assert config.get_option("option2") == old_value_2

    # Make sure that config gets reset when exceptions occur
    try:
        with config.with_options(option1="lkjasdlfkjasdf"):
            raise RuntimeError("something went wrong!")
    except RuntimeError:
        pass
    assert config.get_option("option1") == old_value_1


def test_invalid_option_warnings(config):
    error_msg = "Invalid option specified: invalid_option"

    with pytest.raises(KeyError, match=error_msg):
        config.get_option("invalid_option")

    with pytest.raises(KeyError, match=error_msg):
        config.set_option("invalid_option", "updated_value")

    with pytest.raises(KeyError, match=error_msg):
        config.reset_option("invalid_option")


def test_repr(config, defaults):
    repr_output = repr(config)
    assert "Woodwork Global Config Settings" in repr_output
    for key, value in defaults.items():
        assert f"{key}: {value}" in repr_output



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import shutil
import sys
from pathlib import Path

from sphinx.ext.autodoc import Documenter, MethodDocumenter

import woodwork

path = os.path.join("..", "..")
sys.path.insert(0, os.path.abspath(path))


# -- Project information -----------------------------------------------------

project = "Woodwork"
copyright = "2020, Alteryx, Inc."
author = "Alteryx, Inc."

# The short X.Y version
version = woodwork.__version__
# The full version, including alpha/beta/rc tags
release = woodwork.__version__


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "nbsphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.extlinks",
    "sphinx.ext.inheritance_diagram",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
    "sphinx_inline_tabs",
    "sphinx_copybutton",
    "myst_parser",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The main toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["**.ipynb_checkpoints"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "pygment_light_style": "tango",
    "pygment_dark_style": "native",
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/alteryx/woodwork",
            "icon": "fab fa-github-square",
            "type": "fontawesome",
        },
        {
            "name": "Twitter",
            "url": "https://twitter.com/AlteryxOSS",
            "icon": "fab fa-twitter-square",
            "type": "fontawesome",
        },
        {
            "name": "Slack",
            "url": "https://join.slack.com/t/alteryx-oss/shared_invite/zt-182tyvuxv-NzIn6eiCEf8TBziuKp0bNA",
            "icon": "fab fa-slack",
            "type": "fontawesome",
        },
    ],
    "collapse_navigation": False,
    "navigation_depth": 2,
}

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "images/woodwork.svg"

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "images/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "Woodworkdoc"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "Woodwork.tex",
        "Woodwork Documentation",
        "Alteryx, Inc.",
        "manual",
    ),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        master_doc,
        "woodwork",
        "Woodwork Documentation",
        [author],
        1,
    ),
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "Woodwork",
        "Woodwork Documentation",
        author,
        "Woodwork",
        "One line description of project.",
        "Miscellaneous",
    ),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]

# -- Options for Markdown files ----------------------------------------------

myst_admonition_enable = True
myst_deflist_enable = True
myst_heading_anchors = 3

# -- Options for Sphinx Copy Button ------------------------------------------

copybutton_prompt_text = "myinputprompt"
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# -- Extension configuration -------------------------------------------------

# If woodwork is open-sourced: replace github specific style.css
extlinks = {
    "issue": ("https://github.com/alteryx/woodwork/issues/%s", "GH#%s"),
    "pr": ("https://github.com/alteryx/woodwork/pull/%s", "GH#%s"),
    "user": ("https://github.com/%s", "@%s"),
}

autosummary_generate = ["api_reference.rst"]
templates_path = ["_templates"]

html_show_sphinx = False
nbsphinx_execute = "always"
nbsphinx_timeout = (
    600  # sphinx defaults each cell to 30 seconds so we need to override here
)

inheritance_graph_attrs = dict(
    rankdir="LR",
    size='"1000, 333"',
    fontsize=30,
    labelfontsize=30,
    ratio="compress",
    dpi=960,
)


class AccessorLevelDocumenter(Documenter):
    """
    Documenter subclass for objects on accessor level (methods, attributes).

    Referenced pandas-sphinx-theme (https://github.com/pandas-dev/pandas-sphinx-theme)
    and sphinx-doc (https://github.com/sphinx-doc/sphinx/blob/8c7faed6fcbc6b7d40f497698cb80fc10aee1ab3/sphinx/ext/autodoc/__init__.py#L846)
    """

    def resolve_name(self, modname, parents, path, base):
        modname = "woodwork"
        mod_cls = path.rstrip(".")
        mod_cls = mod_cls.split(".")
        return modname, mod_cls + [base]


class AccessorCallableDocumenter(AccessorLevelDocumenter, MethodDocumenter):
    """
    This documenter lets us removes .__call__ from the method signature for
    callable accessors like Series.plot
    """

    objtype = "accessorcallable"
    directivetype = "method"

    # lower than MethodDocumenter; otherwise the doc build prints warnings
    priority = 0.5

    def format_name(self):
        return MethodDocumenter.format_name(self).rstrip(".__call__")


class AccessorMethodDocumenter(AccessorLevelDocumenter, MethodDocumenter):
    objtype = "accessormethod"
    directivetype = "method"

    # lower than MethodDocumenter so this is not chosen for normal methods
    priority = 0.6


def setup(app):
    home_dir = os.environ.get("HOME", "/")
    ipython_p = Path(home_dir + "/.ipython/profile_default/startup")
    ipython_p.mkdir(parents=True, exist_ok=True)
    file_p = os.path.abspath(os.path.dirname(__file__))
    shutil.copy(
        file_p + "/set-headers.py",
        home_dir + "/.ipython/profile_default/startup",
    )
    app.add_js_file(
        "https://cdnjs.cloudflare.com/ajax/libs/require.js/2.1.10/require.min.js",
    )
    app.add_css_file("style.css")
    app.add_autodocumenter(AccessorCallableDocumenter)
    app.add_autodocumenter(AccessorMethodDocumenter)
